#!/usr/bin/env python
# coding=utf-8
# Copyright (c) 2018 UT-BATTELLE, LLC
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""General tools for working with ensembles."""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from evv4esm import pf_color_picker

def monthly_to_annual_avg(var_data, cal='ignore'):
    if len(var_data) != 12:
        raise ValueError('Error! There are 12 months in a year; '
                         'you passed in {} monthly averages.'.format(len(var_data)))

    if cal == 'ignore':
        # weight each month equally
        avg = np.average(var_data)
    else:
        # TODO: more advanced calendar handling
        raise NotImplementedError
    return avg


def prob_plot(test, ref, n_q, img_file, test_name='Test', ref_name='Ref.',
              thing='annual global averages', pf=None):
    # NOTE: Following the methods described in
    #       https://stackoverflow.com/questions/43285752
    #       to create the Q-Q and P-P plots
    q = np.linspace(0, 100, n_q + 1)
    all_ = np.concatenate((test, ref))
    min_ = np.min(all_)
    max_ = np.max(all_)

    fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, figsize=(10, 10))
    plt.rc('font', family='serif')

    ax1.set_title('Q-Q Plot')
    ax1.set_xlabel('{} pdf'.format(ref_name))
    ax1.set_ylabel('{} pdf'.format(test_name))

    # NOTE: Axis switched here from Q-Q plot because cdf reflects about the 1-1 line
    ax2.set_title('P-P Plot')
    ax2.set_xlabel('{} cdf'.format(test_name))
    ax2.set_ylabel('{} cdf'.format(ref_name))

    ax3.set_title('{} pdf'.format(ref_name))
    ax3.set_xlabel('Unity-based normalization of {}'.format(thing))
    ax3.set_ylabel('Frequency')

    ax4.set_title('{} pdf'.format(test_name))
    ax4.set_xlabel('Unity-based normalization of {}'.format(thing))
    ax4.set_ylabel('Frequency')

    norm_rng = [0.0, 1.0]
    ax1.plot(norm_rng, norm_rng, 'gray', zorder=1)
    ax1.set_xlim(tuple(norm_rng))
    ax1.set_ylim(tuple(norm_rng))
    ax1.autoscale()

    ax2.plot(norm_rng, norm_rng, 'gray', zorder=1)
    ax2.set_xlim(tuple(norm_rng))
    ax2.set_ylim(tuple(norm_rng))
    ax2.autoscale()

    ax3.set_xlim(tuple(norm_rng))
    ax3.autoscale()

    ax4.set_xlim(tuple(norm_rng))
    ax4.autoscale()

    # NOTE: Produce unity-based normalization of data for the Q-Q plots because
    #       matplotlib can't handle small absolute values or data ranges. See
    #          https://github.com/matplotlib/matplotlib/issues/6015
    if not np.allclose(min_, max_, atol=np.finfo(max_).eps):
        norm1 = (ref - min_) / (max_ - min_)
        norm2 = (test - min_) / (max_ - min_)

        ax1.scatter(np.percentile(norm1, q), np.percentile(norm2, q),
                    color=pf_color_picker.get(pf, '#1F77B4'), zorder=2)
        ax3.hist(norm1, bins=n_q, color=pf_color_picker.get(pf, '#1F77B4'))
        ax4.hist(norm2, bins=n_q, color=pf_color_picker.get(pf, '#1F77B4'))

    # bin both series into equal bins and get cumulative counts for each bin
    bnds = np.linspace(min_, max_, n_q)
    if not np.allclose(bnds, bnds[0], atol=np.finfo(bnds[0]).eps):
        ppxb = pd.cut(ref, bnds)
        ppyb = pd.cut(test, bnds)

        ppxh = ppxb.value_counts().sort_index(ascending=True) / len(ppxb)
        ppyh = ppyb.value_counts().sort_index(ascending=True) / len(ppyb)

        ppxh = np.cumsum(ppxh)
        ppyh = np.cumsum(ppyh)

        ax2.scatter(ppyh.values, ppxh.values,
                    color=pf_color_picker.get(pf, '#1F77B4'), zorder=2)

    plt.tight_layout()
    plt.savefig(img_file, bbox_inches='tight')

    plt.close(fig)

    return img_file



# coding=utf-8
# Copyright (c) 2018 UT-BATTELLE, LLC
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import, division, print_function, unicode_literals

__version_info__ = (0, 2, 0)
__version__ = '.'.join(str(vi) for vi in __version_info__)

PASS_COLOR = '#389933'
L_PASS_COLOR = '#93DA90'

FAIL_COLOR = '#BF3F46'
L_FAIL_COLOR = '#E68388'

pf_color_picker = {'Pass': PASS_COLOR, 'pass': PASS_COLOR,
                   'Accept': PASS_COLOR, 'accept': PASS_COLOR,
                   'Fail': FAIL_COLOR, 'fail': FAIL_COLOR,
                   'Reject': FAIL_COLOR, 'reject': FAIL_COLOR}

light_pf_color_picker = {'Pass': L_PASS_COLOR, 'pass': L_PASS_COLOR,
                         'Accept': L_PASS_COLOR, 'accept': L_PASS_COLOR,
                         PASS_COLOR: L_PASS_COLOR,
                         'Fail': L_FAIL_COLOR, 'fail': L_FAIL_COLOR,
                         'Reject': L_FAIL_COLOR, 'reject': L_FAIL_COLOR,
                         FAIL_COLOR: L_FAIL_COLOR}



#!/usr/bin/env python
# coding=utf-8
# Copyright (c) 2015,2016, UT-BATTELLE, LLC
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""The Time Step Convergence Test:
This tests the null hypothesis that the convergence of the time stepping error
for a set of key atmospheric variables is the same for a reference ensemble and
a test ensemble. Both the reference and test ensemble are generated with a
two-second time step, and for each variable the RMSD between each ensemble and
a truth ensemble, generated with a one-second time step, is calculated. RMSD is
calculated globally and over two domains, the land and the ocean. The land/ocean
domains contain just the atmosphere points that are over land/ocean cells.

At each 10 second interval during the 10 minute long simulations, the difference
in the reference and test RMSDs for each variable, each ensemble member, and each
domain are calculated and these ΔRMSDs should be zero for identical climates. A
one sided (due to self convergence) Student's T Test is used to test the null
hypothesis that the ensemble mean ΔRMSD is statistically zero at the {}%
confidence level. A rejection of the null hypothesis (mean ΔRMSD is not
statistically zero) at any time step for any variable will cause this test to
fail.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import argparse
import glob
import os
from collections import OrderedDict
from itertools import groupby
from pprint import pprint

import numpy as np
import pandas as pd
import matplotlib.ticker as tkr
import matplotlib.pyplot as plt
from scipy import stats
from netCDF4 import Dataset

import livvkit
from livvkit.util import elements as el
from livvkit.util import functions as fn
from livvkit.util.LIVVDict import LIVVDict

from evv4esm.ensembles import e3sm
from evv4esm.utils import bib2html
from evv4esm import pf_color_picker, light_pf_color_picker


def parse_args(args=None):
    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('-c', '--config',
                        type=fn.read_json,
                        default='test/tsc_pc0101123.json',
                        help='A JSON config file containing a `tsc` dictionary defining ' +
                             'the options.')

    args = parser.parse_args(args)
    name = args.config.keys()[0]
    config = args.config[name]

    return name, config


def run(name, config, print_details=False):
    """
    Runs the extension.

    Args:
        name: The name of the extension
        config: The test's config dictionary
        print_details: Whether to print the analysis details to stdout
                       (default: False)

    Returns:
       A LIVVkit page element containing the LIVVkit elements to display on a webpage
    """

    # FIXME: move into a config to NameSpace function
    test_args = OrderedDict([(k.replace('-', '_'), v) for k, v in config.items()])
    test_args = argparse.Namespace(**test_args)

    test_args.img_dir = os.path.join(livvkit.output_dir, 'validation', 'imgs', name)
    fn.mkdir_p(test_args.img_dir)
    details, img_gal = main(test_args)

    if print_details:
        _print_details(details)

    domain_headers = ['Null hypothesis', 'T test (t, P)']
    global_tbl_el = {'Type': 'V-H Table',
                     'Title': 'Validation',
                     'TableTitle': 'Analyzed variables',
                     'Headers': domain_headers,
                     'Data': details['global'],
                     }
    land_tbl_el = {'Type': 'V-H Table',
                   'Title': 'Validation',
                   'TableTitle': 'Analyzed variables',
                   'Headers': domain_headers,
                   'Data': details['land'],
                   }
    ocean_tbl_el = {'Type': 'V-H Table',
                    'Title': 'Validation',
                    'TableTitle': 'Analyzed variables',
                    'Headers': domain_headers,
                    'Data': details['ocean'],
                    }
    bib_html = bib2html(os.path.join(os.path.dirname(__file__), 'tsc.bib'))
    tab_list = [el.tab('Figures', element_list=[img_gal]),
                el.tab('Global_details', element_list=[global_tbl_el]),
                el.tab('Land_details', element_list=[land_tbl_el]),
                el.tab('Ocean_details', element_list=[ocean_tbl_el]),
                el.tab('References', element_list=[el.html(bib_html)])]

    results = {'Type': 'Table',
               'Title': 'Results',
               'Headers': ['Test status', 'Global', 'Land', 'Ocean', 'Ensembles'],
               'Data': {'Test status': details['overall'],
                        'Global': details['domains']['delta_l2_global'],
                        'Land': details['domains']['delta_l2_land'],
                        'Ocean': details['domains']['delta_l2_ocean'],
                        'Ensembles': 'statistically identical' if details['overall'] == 'Pass' else 'statistically different',
                        }
               }

    # FIXME: Put into a ___ function
    doc_text = __doc__.format((1 - test_args.p_threshold) * 100).replace('\n\n', '<br><br>')
    page = el.page(name, doc_text, element_list=[results], tab_list=tab_list)
    return page


def main(args):
    if args.test_case == args.ref_case:
        args.test_case += '1'
        args.ref_case += '2'

    file_search_glob = '{d}/*.cam_????.h0.0001-01-01-?????.nc.{s}'
    truth_ens = {instance: list(files) for instance, files in groupby(
            sorted(glob.glob(file_search_glob.format(d=args.ref_dir, s='DT0001'))),
            key=lambda f: e3sm.component_file_instance('cam', f))}

    ref_ens = {instance: list(files) for instance, files in groupby(
            sorted(glob.glob(file_search_glob.format(d=args.ref_dir, s='DT0002'))),
            key=lambda f: e3sm.component_file_instance('cam', f))}

    test_ens = {instance: list(files) for instance, files in groupby(
            sorted(glob.glob(file_search_glob.format(d=args.test_dir, s='DT0002'))),
            key=lambda f: e3sm.component_file_instance('cam', f))}

    # So, we want a pandas dataframe that will have the columns :
    #     (test/ref, ensemble, seconds, l2_global, l2_land, l2_ocean)
    # But, building a pandas dataframe row by row is sloooow, so append a list of list and convert
    data = []
    for instance, truth_files in truth_ens.items():
        times = [int(e3sm.file_date_str(ff, style='full').split('-')[-1]) for ff in truth_files]
        for tt, time in enumerate(times):
            with Dataset(truth_ens[instance][tt]) as truth, \
                    Dataset(ref_ens[instance][tt]) as ref, \
                    Dataset(test_ens[instance][tt]) as test:

                truth_plt, truth_ps = pressure_layer_thickness(truth)
                ref_plt, ref_ps = pressure_layer_thickness(ref)
                test_plt, test_ps = pressure_layer_thickness(test)

                ref_avg_plt = (truth_plt + ref_plt) / 2.0
                test_avg_plt = (truth_plt + test_plt) / 2.0

                ref_avg_ps = np.expand_dims((truth_ps + ref_ps) / 2.0, 0)
                test_avg_ps = np.expand_dims((truth_ps + test_ps) / 2.0, 0)

                ref_area = np.expand_dims(ref.variables['area'][...], 0)
                test_area = np.expand_dims(test.variables['area'][...], 0)

                ref_landfrac = np.expand_dims(ref.variables['LANDFRAC'][0, ...], 0)
                test_landfrac = np.expand_dims(test.variables['LANDFRAC'][0, ...], 0)

                for var in args.variables:
                    truth_var = truth.variables[var][0, ...]
                    ref_var = ref.variables[var][0, ...]
                    test_var = test.variables[var][0, ...]

                    ref_int = (ref_var - truth_var) ** 2 * ref_avg_plt
                    test_int = (test_var - truth_var) ** 2 * test_avg_plt

                    ref_global_l2 = np.sqrt((ref_int * ref_area).sum()
                                            / (ref_avg_ps * ref_area).sum())
                    ref_land_l2 = np.sqrt((ref_int * (ref_area * ref_landfrac)).sum()
                                          / (ref_avg_ps * (ref_area * ref_landfrac)).sum())
                    ref_ocean_l2 = np.sqrt((ref_int * (ref_area * (1 - ref_landfrac))).sum()
                                           / (ref_ps * (ref_area * (1 - ref_landfrac))).sum())

                    test_global_l2 = np.sqrt((test_int * test_area).sum()
                                             / (test_avg_ps * test_area).sum())
                    test_land_l2 = np.sqrt((test_int * (test_area * test_landfrac)).sum()
                                           / (test_avg_ps * (test_area * test_landfrac)).sum())
                    test_ocean_l2 = np.sqrt((test_int * (test_area * (1 - test_landfrac))).sum()
                                            / (test_ps * (test_area * (1 - test_landfrac))).sum())

                    # (case, ensemble, seconds, var, l2_global, l2_land, l2_ocean)
                    data.append([args.ref_case, instance, time, var,
                                 ref_global_l2, ref_land_l2, ref_ocean_l2])
                    data.append([args.test_case, instance, time, var,
                                 test_global_l2, test_land_l2, test_ocean_l2])

    tsc_df = pd.DataFrame(data, columns=['case', 'instance', 'seconds', 'variable',
                                         'l2_global', 'l2_land', 'l2_ocean'])

    # NOTE: This is what we're actually going to apply the one-sided t test to
    test_columns = ['l2_global', 'l2_land', 'l2_ocean']
    delta_columns = ['delta_' + t for t in test_columns]
    ref_columns = ['ref_' + t for t in test_columns]
    delta_rmsd = tsc_df[tsc_df['case'] == args.ref_case].copy()
    delta_rmsd[delta_columns] = tsc_df[tsc_df['case'] == args.test_case][test_columns].values \
                                - delta_rmsd[test_columns]
    delta_rmsd.rename(columns={t: r for t, r in zip(test_columns, ref_columns)}, inplace=True)

    # NOTE: This mess is because we want to "normalize" delta l2 by the mean
    # _reference_ l2 for each instance and variable.
    delta_rmsd[['norm_' + d for d in delta_columns]] = \
        delta_rmsd.groupby(['instance', 'variable']).apply(
            lambda g: g[delta_columns] / g[ref_columns].mean().values)

    testee = delta_rmsd.query(' seconds >= @args.time_slice[0] & seconds <= @args.time_slice[-1]')
    ttest = testee.groupby(['seconds', 'variable']).agg(stats.ttest_1samp, popmean=0.0).drop(columns='instance')

    # H0: enemble_mean_ΔRMSD_{t,var} is (statistically) zero and therefore, the simulations are identical
    null_hypothesis = ttest.applymap(lambda x: 'Reject' if x[1] < args.p_threshold else 'Accept')

    domains = null_hypothesis.applymap(lambda x: x == 'Reject').any().transform(
            lambda x: 'Fail' if x is True else 'Pass')
    overall = 'Fail' if domains.apply(lambda x: x == 'Fail').any() else 'Pass'

    ttest.reset_index(inplace=True)
    null_hypothesis.reset_index(inplace=True)

    land_data = LIVVDict()
    ocean_data = LIVVDict()
    global_data = LIVVDict()
    for sec in ttest['seconds'].unique():
        for var in ttest['variable'].unique():
            sec_str = 'Time: {:03d}s'.format(sec)
            t_data = ttest.loc[(ttest['seconds'] == sec) & (ttest['variable'] == var)]
            h0_data = null_hypothesis.loc[(null_hypothesis['seconds'] == sec) & (null_hypothesis['variable'] == var)]
            global_data[sec_str][var]['Null hypothesis'] = h0_data['delta_l2_global'].values[0]
            land_data[sec_str][var]['Null hypothesis'] = h0_data['delta_l2_land'].values[0]
            ocean_data[sec_str][var]['Null hypothesis'] = h0_data['delta_l2_ocean'].values[0]

            global_data[sec_str][var]['T test (t, P)'] = '({:.3f}, {:.4f})'.format(*t_data['delta_l2_global'].values[0])
            land_data[sec_str][var]['T test (t, P)'] = '({:.3f}, {:.4f})'.format(*t_data['delta_l2_land'].values[0])
            ocean_data[sec_str][var]['T test (t, P)'] = '({:.3f}, {:.4f})'.format(*t_data['delta_l2_ocean'].values[0])

    details = {'ocean': ocean_data, 'land': land_data, 'global': global_data,
               'domains': domains, 'overall': overall}

    fail_timeline_img = os.path.relpath(os.path.join(args.img_dir, 'failing_timeline.png'), os.getcwd())
    pmin_timeline_img = os.path.relpath(os.path.join(args.img_dir, 'pmin_timeline.png'), os.getcwd())
    rmsd_img_format = os.path.relpath(os.path.join(args.img_dir, '{}_rmsd_{{}}s.png'), os.getcwd())

    img_list = [plot_failing_variables(args, null_hypothesis, fail_timeline_img),
                plot_pmin(args, ttest, pmin_timeline_img)]
    e_img_list = errorbars_delta_rmsd(args, delta_rmsd[delta_rmsd['seconds'].isin(args.inspect_times)],
                                      null_hypothesis, rmsd_img_format.format('distribution'))

    b_img_list = boxplot_delta_rmsd(args, delta_rmsd[delta_rmsd['seconds'].isin(args.inspect_times)],
                                    null_hypothesis, rmsd_img_format.format('boxplot'))

    img_list.extend([img for pair in zip(e_img_list, b_img_list) for img in pair])
    img_gallery = el.gallery('Time step convergence', img_list)

    return details, img_gallery


def pressure_layer_thickness(dataset):
    p0 = dataset.variables['P0'][...]
    ps = dataset.variables['PS'][0, ...]
    hyai = dataset.variables['hyai'][...]
    hybi = dataset.variables['hybi'][...]

    da = hyai[1:] - hyai[0:-1]
    db = hybi[1:] - hybi[0:-1]

    dp = np.expand_dims(da * p0, 1) + (np.expand_dims(db, 1) * np.expand_dims(ps, 0))
    return dp, ps


def plot_failing_variables(args, null_hypothesis, img_file):
    null_hypothesis[['n_fail_land', 'n_fail_ocean']] = \
        null_hypothesis[['delta_l2_land', 'delta_l2_ocean']].transform(lambda x: x == 'Reject').astype('int')

    pdata = null_hypothesis[['seconds', 'n_fail_land', 'n_fail_ocean']].groupby('seconds').sum().sum(axis=1)

    fig, ax = plt.subplots(figsize=(10, 8))
    plt.rc('font', family='serif')

    pdata.plot(linestyle='-', marker='o', color=pf_color_picker.get('fail'))
    ax.set_ybound(0, 20)
    ax.set_yticks(np.arange(0, 24, 4))
    ax.set_yticks(np.arange(0, 21, 1), minor=True)

    ax.set_ylabel('Number of failing variables')
    ax.set_xlabel('Integration time (s)')

    plt.tight_layout()
    plt.savefig(img_file, bbox_inches='tight')
    plt.close(fig)

    img_caption = 'The number of failing variables across both domains (land and ' \
                  'ocean) as a function of model integration time.'
    img_link = os.path.join(os.path.basename(args.img_dir), os.path.basename(img_file))
    img = el.image('Timeline of failing variables', img_caption, img_link, height=300)
    return img


def plot_pmin(args, ttest, img_file):
    ttest[['p_land', 'p_ocean']] = ttest[['delta_l2_land', 'delta_l2_ocean']].applymap(lambda x: x[1])
    pdata = ttest[['seconds', 'p_land', 'p_ocean']].groupby(['seconds']).min().min(axis=1) * 100  # to %

    fig, ax = plt.subplots(figsize=(10, 8))
    plt.rc('font', family='serif')

    fails = pdata[pdata.lt(args.p_threshold * 100)]
    passes = pdata[pdata.ge(args.p_threshold * 100)]

    if passes.empty:
        fails.plot(logy=True, linestyle='-', marker='o', color=pf_color_picker.get('fail'))
    elif fails.empty:
        passes.plot(logy=True, linestyle='-', marker='o', color=pf_color_picker.get('pass'))
    else:
        first_fail = fails.index[0]
        pdata.loc[:first_fail].plot(logy=True, linestyle='-', marker='o', color=pf_color_picker.get('pass'))
        pdata.loc[first_fail:].plot(logy=True, linestyle='-', marker='o', color=pf_color_picker.get('fail'))

    ax.plot(args.time_slice, [0.5, 0.5], 'k--')
    ax.text(np.mean(args.time_slice), 10 ** -1, 'Fail', fontsize=15, color=pf_color_picker.get('fail'),
            horizontalalignment='center')
    ax.text(np.mean(args.time_slice), 10 ** 0, 'Pass', fontsize=15, color=pf_color_picker.get('pass'),
            horizontalalignment='center')

    ax.set_ybound(100, 10 ** -15)
    locmaj = tkr.LogLocator(numticks=18)
    ax.yaxis.set_major_locator(locmaj)
    locmin = tkr.LogLocator(subs=(0.2, 0.4, 0.6, 0.8), numticks=18)
    ax.yaxis.set_minor_locator(locmin)
    ax.yaxis.set_minor_formatter(tkr.NullFormatter())
    ax.set_ylabel('P_{min} (%)')
    ax.set_xlabel('Integration time (s)')

    plt.tight_layout()
    plt.savefig(img_file, bbox_inches='tight')
    plt.close(fig)

    img_caption = 'The minimum P value of all variables in both domains (land and ' \
                  'ocean) as a function of model integration time plotted with ' \
                  'a logarithmic y-scale. The dashed grey line indicates the' \
                  'threshold for assigning an overall "pass" or "fail" to a test' \
                  'ensemble, see Wan et al. (2017) eqn. 8.'
    img_link = os.path.join(os.path.basename(args.img_dir), os.path.basename(img_file))
    img = el.image('Timeline of P_{min}', img_caption, img_link, height=300)
    return img


def boxplot_delta_rmsd(args, delta_rmsd, null_hypothesis, img_file_format):
    img_list = []
    columns = ['instance', 'variable', 'norm_delta_l2_land', 'norm_delta_l2_ocean']
    for time in delta_rmsd['seconds'].unique():
        img_file = img_file_format.format(time)
        pdata = delta_rmsd[delta_rmsd['seconds'] == time][columns]

        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(10, 8), sharex='all')
        plt.rc('font', family='serif')

        bp1 = pdata.boxplot(column='norm_delta_l2_land', by='variable', ax=ax1, vert=False, grid=False,
                            patch_artist=True, return_type='dict', showmeans=True,
                            meanprops={'markerfacecolor': 'black',
                                       'markeredgecolor': 'black',
                                       'marker': '.'})
        bp2 = pdata.boxplot(column='norm_delta_l2_ocean', by='variable', ax=ax2, vert=False, grid=False,
                            patch_artist=True, return_type='dict', showmeans=True,
                            meanprops={'markerfacecolor': 'black',
                                       'markeredgecolor': 'black',
                                       'marker': '.'})

        for box1, box2, var1, var2 in zip(bp1['norm_delta_l2_land']['boxes'],
                                          bp2['norm_delta_l2_ocean']['boxes'],
                                          list(ax1.get_yticklabels()),
                                          list(ax2.get_yticklabels())):

            land_var_color = pf_color_picker[
                null_hypothesis[(null_hypothesis['seconds'] == time)
                                & (null_hypothesis['variable'] == var1.get_text())]['delta_l2_land'].values[0]]
            ocean_var_color = pf_color_picker[
                null_hypothesis[(null_hypothesis['seconds'] == time)
                                & (null_hypothesis['variable'] == var2.get_text())]['delta_l2_ocean'].values[0]]

            var1.set_color(land_var_color)
            box1.set_color(land_var_color)
            box1.set_alpha(0.5)

            var2.set_color(ocean_var_color)
            box2.set_color(ocean_var_color)
            box2.set_alpha(0.5)

        for artist in ['fliers', 'medians', 'means', 'whiskers', 'caps']:
            for a1, a2 in zip(bp1['norm_delta_l2_land'][artist],
                              bp2['norm_delta_l2_ocean'][artist]):
                a1.set_color('black')
                a2.set_color('black')

        st = fig.suptitle("Ensemble ΔRMSD at t={}s".format(time), size=16)

        ax1.set_xlabel('Normalized ensemble ΔRMSD')
        ax1.set_title('Land')

        ax2.set_xlabel('Normalized ensemble ΔRMSD')
        ax2.set_title('Ocean')

        plt.savefig(img_file, bbox_extra_artists=[st], bbox_inches='tight')
        plt.close(fig)

        img_caption = 'Boxplot of ensemble ΔRMSD at t={}s, where the dots show ' \
                      'the ensemble mean, ther verical lines show the ensemble median, ' \
                      'the filled boxes cover the interquartile range (IQR), and ' \
                      'the whiskers cover 1.5*IQR. All values here have been ' \
                      'normalized by the mean RMSD of the reference ensemble. ' \
                      'Red and blue indicate fail and pass, respectively, ' \
                      'according to the criterion described in ' \
                      'Wan et al. (2017), eq. 6.'.format(time)
        img_link = os.path.join(os.path.basename(args.img_dir), os.path.basename(img_file))
        img_list.append(el.image('Boxplot of normalized ensemble ΔRMSD at {}s'.format(time),
                                 img_caption, img_link, height=300))
    return img_list


def errorbars_delta_rmsd(args, delta_rmsd, null_hypothesis, img_file_format):
    img_list = []
    columns = ['instance', 'variable', 'norm_delta_l2_land', 'norm_delta_l2_ocean']

    N = len(delta_rmsd['instance'].unique()) - 1
    tval_crit = stats.t.ppf(1 - args.p_threshold, df=N - 1)

    # NOTE: σ/√N is the t-test scaling parameter, see:
    #          https://en.wikipedia.org/wiki/Student%27s_t-test
    #       In the Wan et al. (2017) fig. 5, and the corresponding article text,
    #       indicates that these plots show ±2σ with the filled boxes, BUT they
    #       actually show ±2σ/√N (verified in old NCL code as well).
    scale_std = 1/np.sqrt(N)

    for time in delta_rmsd['seconds'].unique():
        img_file = img_file_format.format(time)
        pdata = delta_rmsd[delta_rmsd['seconds'] == time][columns].groupby('variable').agg(
                ['mean', np.std])

        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(10, 8), sharex='all')
        plt.rc('font', family='serif')

        yvals = pdata.reset_index().index.values + 1
        ylims = [yvals[0] - 0.5, yvals[-1] + 0.5]
        land_colors = [pf_color_picker[
                           null_hypothesis[(null_hypothesis['seconds'] == time)
                                           & (null_hypothesis['variable'] == var)]['delta_l2_land'].values[0]]
                       for var in pdata.index.values]
        l_land_colors = [light_pf_color_picker[lc] for lc in land_colors]
        ocean_colors = [pf_color_picker[
                           null_hypothesis[(null_hypothesis['seconds'] == time)
                                           & (null_hypothesis['variable'] == var)]['delta_l2_ocean'].values[0]]
                        for var in pdata.index.values]
        l_ocean_colors = [light_pf_color_picker[oc] for oc in ocean_colors]

        et1 = ax1.errorbar(pdata['norm_delta_l2_land']['mean'].values, yvals,
                           xerr=np.stack([tval_crit * scale_std * pdata['norm_delta_l2_land']['std'].values,
                                          2 * scale_std * pdata['norm_delta_l2_land']['std'].values]),
                           fmt='k.', elinewidth=20, ecolor='lightblue')
        es1 = ax1.errorbar(pdata['norm_delta_l2_land']['mean'].values, yvals,
                           xerr=2 * scale_std * pdata['norm_delta_l2_land']['std'].values,
                           fmt='k.', elinewidth=20)

        et2 = ax2.errorbar(pdata['norm_delta_l2_ocean']['mean'].values, yvals,
                           xerr=np.stack([tval_crit * scale_std * pdata['norm_delta_l2_ocean']['std'].values,
                                          2 * scale_std * pdata['norm_delta_l2_ocean']['std'].values]),
                           fmt='k.', elinewidth=20, ecolor='lightblue')
        es2 = ax2.errorbar(pdata['norm_delta_l2_ocean']['mean'].values, yvals,
                           xerr=2 * scale_std * pdata['norm_delta_l2_ocean']['std'].values,
                           fmt='k.', elinewidth=20)

        _, _, (elines1,) = et1.lines
        elines1.set_color(l_land_colors)
        elines1.set_alpha(0.5)

        _, _, (elines1,) = es1.lines
        elines1.set_color(land_colors)
        elines1.set_alpha(0.5)

        ax1.plot([0, 0], ylims, 'k-', alpha=0.5)
        ax1.set_ylim(*ylims)
        ax1.set_yticks(yvals)
        ax1.set_yticklabels(pdata.index.values)

        _, _, (elines2,) = et2.lines
        elines2.set_color(l_ocean_colors)
        elines2.set_alpha(0.5)

        _, _, (elines2,) = es2.lines
        elines2.set_color(ocean_colors)
        elines2.set_alpha(0.5)

        ax2.plot([0, 0], ylims, 'k-', alpha=0.5)
        ax2.set_ylim(*ylims)
        ax2.set_yticks(yvals)
        ax2.set_yticklabels(pdata.index.values)

        for ii, var1, var2 in zip(yvals - 1,
                                  list(ax1.get_yticklabels()),
                                  list(ax2.get_yticklabels())):

            var1.set_color(land_colors[ii])
            var2.set_color(ocean_colors[ii])

        st = fig.suptitle("Ensemble ΔRMSD at t={}s".format(time), size=16)

        ax1.set_xlabel('Normalized ensemble ΔRMSD')
        ax1.set_title('Land')

        ax2.set_xlabel('Normalized ensemble ΔRMSD')
        ax2.set_title('Ocean')

        plt.savefig(img_file, bbox_extra_artists=[st], bbox_inches='tight')
        plt.close(fig)

        img_caption = 'Ensemble-mean ΔRMSD at t={}s (dots) and the ±2σ/√N range of ' \
                      'the mean (dark filled boxes), where σ denotes the standard ' \
                      'deviation, N the ensemble size, and σ/√N is the t-test scaling ' \
                      'parameter. The left end of the light filled boxes shows the ' \
                      'threshold value corresponding to the critical P {}% in the ' \
                      'one-sided t-test. All values here have been ' \
                      'normalized by the mean RMSD of the reference ensemble. ' \
                      'Red and blue indicate fail and pass, respectively, ' \
                      'according to the criterion described in ' \
                      'Wan et al. (2017), eq. 6.'.format(time, args.p_threshold * 100)
        img_link = os.path.join(os.path.basename(args.img_dir), os.path.basename(img_file))
        img_list.append(el.image('Distribution of the ensemble ΔRMSD at {}s'.format(time),
                                 img_caption, img_link, height=300))
    return img_list


def _print_details(details):
    for set_ in details:
        print('-' * 80)
        print(set_)
        print('-' * 80)
        pprint(details[set_])


def print_summary(summary):
    print('    Time step convergence test: {}'.format(summary['']['Case']))
    print('      Global: {}'.format(summary['']['Global']))
    print('      Land: {}'.format(summary['']['Land']))
    print('      Ocean: {}'.format(summary['']['Ocean']))
    print('      Ensembles: {}'.format(summary['']['Ensembles']))
    print('      Test status: {}\n'.format(summary['']['Test status']))


def summarize_result(results_page):
    summary = {'Case': results_page['Title']}
    for elem in results_page['Data']['Elements']:
        if elem['Type'] == 'Table' and elem['Title'] == 'Results':
               summary['Global'] = elem['Data']['Global']
               summary['Land'] = elem['Data']['Land']
               summary['Ocean'] = elem['Data']['Ocean']
               summary['Ensembles'] = elem['Data']['Ensembles']
               summary['Test status'] = elem['Data']['Test status']
    return {'': summary}


def populate_metadata():
    """
    Generates the metadata responsible for telling the summary what
    is done by this module's run method
    """

    metadata = {'Type': 'ValSummary',
                'Title': 'Validation',
                'TableTitle': 'Time step convergence test',
                'Headers': ['Test status', 'Global', 'Land', 'Ocean', 'Ensembles']}
    return metadata


if __name__ == '__main__':
    test_name, test_config = parse_args()
    run(test_name, test_config, print_details=True)



#!/usr/bin/env python
# coding=utf-8
# Copyright (c) 2018 UT-BATTELLE, LLC
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# 
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import, division, print_function, unicode_literals

import os
import sys
import time
import argparse

import evv4esm
import livvkit
from livvkit.util import options


def parse_args(args=None):
    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('-e', '--extensions',
                        action='store',
                        nargs='+',
                        default=None,
                        help='Specify the location of the JSON configuration files for the extended V&V tests to run.')
  
    parser.add_argument('-o', '--out-dir',
                        default=os.path.join(os.getcwd(), "vv_" + time.strftime("%Y-%m-%d")),
                        help='Location to output the EVV webpages.')

    parser.add_argument('-s', '--serve',
                        nargs='?', type=int, const=8000,
                        help=' '.join(['Start a simple HTTP server for the output website specified',
                                       'by OUT_DIR on port SERVE.'
                                       ])
                        )

    parser.add_argument('--version',
                        action='version',
                        version='EVV {}'.format(evv4esm.__version__),
                        help="Show EVV's version number and exit"
                        )

    args = parser.parse_args(args)

    if args.extensions:
        options.parse_args(['-V']+args.extensions + ['-o', args.out_dir])
    
    from evv4esm import resources
    args.livv_resource_dir = livvkit.resource_dir
    livvkit.resource_dir = os.sep.join(resources.__path__)
    
    return args 


def main(cl_args=None):
    """ Direct execution. """

    if cl_args is None and len(sys.argv) > 1:
        cl_args = sys.argv[1:]
    args = parse_args(cl_args)

    print("--------------------------------------------------------------------")
    print("                   ______  __      __ __      __                    ")
    print("                  |  ____| \ \    / / \ \    / /                    ")
    print("                  | |__     \ \  / /   \ \  / /                     ")
    print("                  |  __|     \ \/ /     \ \/ /                      ")
    print("                  | |____     \  /       \  /                       ")
    print("                  |______|     \/         \/                        ")
    print("                                                                    ")
    print("    Extended Verification and Validation for Earth System Models    ")
    print("--------------------------------------------------------------------")
    print("")
    print("  Current run: " + livvkit.timestamp)
    print("  User: " + livvkit.user)
    print("  OS Type: " + livvkit.os_type)
    print("  Machine: " + livvkit.machine)
    print("  " + livvkit.comment)

    from livvkit.components import validation
    from livvkit import scheduler
    from livvkit.util import functions
    from livvkit.util import elements

    if args.extensions:
        functions.setup_output(jsd=os.path.join(args.livv_resource_dir, 'js'))
    
        summary_elements = []
        validation_config = {}
        print(" -----------------------------------------------------------------")
        print("   Beginning extensions test suite ")
        print(" -----------------------------------------------------------------")
        print("")
        for conf in livvkit.validation_model_configs:
            validation_config = functions.merge_dicts(validation_config,
                                                      functions.read_json(conf))
        summary_elements.extend(scheduler.run_quiet(validation, validation_config,
                                                    group=False))
        print(" -----------------------------------------------------------------")
        print("   Extensions test suite complete ")
        print(" -----------------------------------------------------------------")
        print("")

        result = elements.page("Summary", "", element_list=summary_elements)
        functions.write_json(result, livvkit.output_dir, "index.json")
        print("-------------------------------------------------------------------")
        print(" Done!  Results can be seen in a web browser at:")
        print("   " + os.path.join(livvkit.output_dir, 'index.html'))
        print("-------------------------------------------------------------------")

    if args.serve:
        try:
            # Python 3
            import http.server as server
            import socketserver as socket
        except ImportError:
            # Python 2
            # noinspection PyPep8Naming
            import SimpleHTTPServer as server
            # noinspection PyPep8Naming
            import SocketServer as socket

        httpd = socket.TCPServer(('', args.serve), server.SimpleHTTPRequestHandler)

        sa = httpd.socket.getsockname()
        print('\nServing HTTP on {host} port {port} (http://{host}:{port}/)'.format(host=sa[0], port=sa[1]))
        print('\nView the generated website by navigating to:')
        print('\n    http://{host}:{port}/{path}/index.html'.format(host=sa[0], port=sa[1],
                                                                    path=os.path.relpath(args.out_dir)
                                                                    ))
        print('\nExit by pressing `ctrl+c` to send a keyboard interrupt.\n')
        try:
            httpd.serve_forever()
        except KeyboardInterrupt:
            print('\nKeyboard interrupt received, exiting.\n')
            sys.exit(0)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding=utf-8
# Copyright (c) 2015,2016, UT-BATTELLE, LLC
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""The Perturbation Growth Test:
This tests the null hypothesis that the reference (n) and modified (m) model
ensembles represent the same atmospheric state after each physics parameterization
is applied within a single time-step using the two-sample (n and m) T-test for equal
averages at a 95% confidence level. Ensembles are generated by repeating the
simulation for many initial conditions, with each initial condition subject to
multiple perturbations.
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import six

import os
import math
import argparse
# import logging

from pprint import pprint
from collections import OrderedDict

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import matplotlib.patches as mpatches

from scipy import stats
from netCDF4 import Dataset

import livvkit
from livvkit.util import elements as el
from livvkit.util import functions as fn

from evv4esm.utils import bib2html

# logger = logging.getLogger(__name__)


def parse_args(args=None):
    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('-c', '--config',
                        type=fn.read_json,
                        default='test/pge_pc0101123.json',
                        help='A JSON config file containing a `pg` dictionary defining ' +
                             'the options.')

    args = parser.parse_args(args)
    name = args.config.keys()[0]
    config = args.config[name]

    return name, config


def _instance2sub(instance_number, total_perturbations):
    """
    Converts an instance number (ii) to initial condition index (ci) and
    perturbation index (pi)  subscripts

    instances use 1-based indexes and vary according to this function:
        ii = ci * len(PERTURBATIONS) + pi + 1
    where both pi and ci use 0-based indexes.
    """
    perturbation_index = (instance_number - 1) % total_perturbations
    initial_condition = (instance_number - 1 - perturbation_index) // total_perturbations
    return initial_condition, perturbation_index


def _sub2instance(initial_condition, perturbation_index, total_perturbations):
    """
    Converts initial condition index (ci) and perturbation index (pi) subscripts
    to an instance number (ii)

    instances use 1-based indexes and vary according to this function:
        ii = ci * len(PERTURBATIONS) + pi + 1
    where both pi and ci use 0-based indexes.
    """
    instance = initial_condition * total_perturbations + perturbation_index + 1
    return instance


def rmse_writer(file_name, rmse, perturbation_names, perturbation_variables, init_file_template):
    """
    Opens and writes a netcdf file for PGE curves
    This function is here purely to avoid duplicate
    codes so that it is easy to maintain code longterm
    """

    with Dataset(file_name, 'w') as nc:
        ninit, nprt_m1, nvars = rmse.shape

        nc.createDimension('ninit', ninit)
        nc.createDimension('nprt', nprt_m1 + 1)
        nc.createDimension('nprt_m1', nprt_m1)
        nc.createDimension('nvars', nvars)

        nc_init_cond = nc.createVariable('init_cond_files', str, 'ninit')
        nc_perturbation = nc.createVariable('perturbation_names', str, 'nprt')
        nc_variables = nc.createVariable('perturbation_variables', str, 'nvars')
        nc_rmse = nc.createVariable('rmse', 'f8', ('ninit', 'nprt_m1', 'nvars'))

        # NOTE: Assignment to netcdf4 variable length string array can be done
        #       via numpy arrays, or in a for loop using integer indices
        nc_perturbation[:] = np.array(perturbation_names)
        nc_variables[:] = np.array(perturbation_variables)
        nc_rmse[:] = rmse[:]

        for icond in range(0, ninit):
            # NOTE: Zero vs One based indexing
            nc_init_cond[icond] = init_file_template.format('cam', 'i', icond+1)


def variables_rmse(ifile_test, ifile_cntl, var_list, var_pefix=''):
    """
    Compute RMSE difference between perturbation and control for a set of
    variables

    Args:
         ifile_test: Path to a NetCDF dataset for a perturbed simulation
         ifile_cntl: Path to a NetCDF dataset for the control simulation
         var_list (list): List of all variables to analyze
         var_pefix: Optional prefix (e.g., t_, qv_) to apply to the variable

    returns:
        rmse (pandas.DataFrame): A dataframe containing the RMSE and maximum
            difference details between the perturbed and control simulation

    """

    # ------------------ARGS-------------------------
    # ifile_test: path of test file
    # ifile_cntl: path of control file
    # var_list  : List of all variables
    # var_pefix: Prefix for var_list (e.g. t_, t_ qv_ etc.)
    # -----------------------------------------------

    with Dataset(ifile_test) as ftest, Dataset(ifile_cntl) as fcntl:
        lat = ftest.variables['lat']
        lon = ftest.variables['lon']

        rmse = pd.DataFrame(columns=('RMSE', 'max diff', 'i', 'j', 'control', 'test', 'lat', 'lon'), index=var_list)

        # reshape for RMSE
        dims = len(ftest.variables[var_pefix + var_list[0]].dimensions)
        if dims == 3:  # see if it is SE grid
            nx, ny = ftest.variables[var_pefix + var_list[0]][0, ...].shape
            nz = 1
        else:
            nx, ny, nz = ftest.variables[var_pefix + var_list[0]][0, ...].shape

        for ivar, vvar in enumerate(var_list):
            var = var_pefix + vvar
            if var in ftest.variables:
                vtest = ftest.variables[var.strip()][0, ...]  # first dimension is time (=0)
                vcntl = fcntl.variables[var.strip()][0, ...]  # first dimension is time (=0)

                vrmse = math.sqrt(((vtest - vcntl)**2).mean()) / np.mean(vcntl)

                diff = abs(vtest[...] - vcntl[...])
                ind_max = np.unravel_index(diff.argmax(), diff.shape)

                rmse.loc[vvar] = (vrmse, diff[ind_max], ind_max[0], ind_max[1],
                                  vcntl[ind_max], vtest[ind_max],
                                  lat[ind_max[1]], lon[ind_max[1]])
    return rmse


def _print_details(details):
    for set_ in details:
        print('-' * 80)
        print(set_)
        print('-' * 80)
        pprint(details[set_])


def main(args):
    
    nvar = len(args.variables)
    nprt = len(args.perturbations)

    # for test cases (new environment etc.)
    # logger.debug("PGN_INFO: Test case comparison...")

    cond_rmse = {}
    for icond in range(args.ninit):
        prt_rmse = {}
        for iprt, prt_name in enumerate(args.perturbations):
            if prt_name == 'woprt':
                continue
            iinst_ctrl = _sub2instance(icond, 0, nprt)
            ifile_ctrl = os.path.join(args.ref_dir,
                                      args.instance_file_template.format('', iinst_ctrl, '_woprt'))
            # logger.debug("PGN_INFO:CNTL_TST:" + ifile_cntl)

            iinst_test = _sub2instance(icond, iprt, nprt)
            ifile_test = os.path.join(args.test_dir,
                                      args.instance_file_template.format(
                                              args.test_case + '.', iinst_test, '_' + prt_name))
            # logger.debug("PGN_INFO:TEST_TST:" + ifile_test)

            prt_rmse[prt_name] = variables_rmse(ifile_test, ifile_ctrl, args.variables, 't_')

        cond_rmse[icond] = pd.concat(prt_rmse)

    rmse = pd.concat(cond_rmse)
    comp_rmse = np.reshape(rmse.RMSE.values, (args.ninit, nprt-1, nvar))

    rmse_writer(os.path.join(args.test_dir, 'comp_cld.nc'),
                comp_rmse, args.perturbations.keys(), args.variables, args.init_file_template)

    details = OrderedDict()
    with Dataset(os.path.join(args.ref_dir, args.pge_cld)) as ref_cld:
        ref_dims = ref_cld.variables['cld_rmse'].shape
        cmp_dims = (args.ninit, nprt - 1, nvar)
        try:
            assert(ref_dims == cmp_dims)
        except AssertionError as e:
            be = BaseException(
                    'PGE curve dimensions (ninit, nptr, nvar) should be the same:\n'
                    '    CLD:{}  COMP:{}'.format(ref_dims, cmp_dims))
            six.raise_from(be, e)

        ref_rmse = ref_cld.variables['cld_rmse'][...]
        details['ref. data'] = ref_rmse

    pge_ends_cld = ref_rmse[:, :, -1]
    pge_ends_comp = comp_rmse[:, :, -1]

    # run the t-test
    pge_ends_cld = pge_ends_cld.flatten()
    pge_ends_comp = pge_ends_comp.flatten()

    t_stat, p_val = stats.ttest_ind(pge_ends_cld, pge_ends_comp)

    if np.isnan((t_stat, p_val)).any() or np.isinf((t_stat, p_val)).any():
        details['T test (t, p)'] = (None, None)
    else:
        details['T test (t, p)'] = '({:.3f}, {:.3f})'.format(t_stat, p_val)

    # logger.warn(" T value:" + str(t_stat))
    # logger.warn(" P value:" + str(p_val))
    crit = 0.05
    if t_stat is None:
        details['h0'] = '-'
    elif p_val < crit:
        details['h0'] = 'reject'
    else:
        details['h0'] = 'accept'

    # logger.debug("PGN_INFO: POST PROCESSING PHASE ENDS")

    details['test data'] = rmse

    ref_max_y = ref_rmse.max(axis=(0, 1)).astype(np.double)
    ref_min_y = ref_rmse.min(axis=(0, 1)).astype(np.double)

    cmp_max_y = comp_rmse.max(axis=(0, 1)).astype(np.double)
    cmp_min_y = comp_rmse.min(axis=(0, 1)).astype(np.double)

    img_file = os.path.relpath(os.path.join(args.img_dir, 'plot_comp.png'), os.getcwd())
    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(10, 8), sharey='all', gridspec_kw={'width_ratios': [3, 1]})
    plt.rc('font', family='serif')

    ax1.semilogy(ref_max_y, color='C0')
    ax1.semilogy(ref_min_y, color='C0')
    ax1.fill_between(range(ref_dims[-1]), ref_min_y, ref_max_y, color='C0', alpha=0.5)

    ax1.semilogy(cmp_max_y, color='C1')
    ax1.semilogy(cmp_min_y, color='C1')
    ax1.fill_between(range(cmp_dims[-1]), cmp_min_y, cmp_max_y, color='C1', alpha=0.5)

    ax1.set_xticks(range(len(args.variables)))
    ax1.set_xticklabels(args.variables, rotation=45, ha='right')
    ax1.set_ylabel('Temperature RMSE (K)')

    patch_list = [mpatches.Patch(color='C0', alpha=0.5, label='Ref.'),
                  mpatches.Patch(color='C1', alpha=0.5, label='Test')]
    ax1.legend(handles=patch_list, loc='upper left')

    scale_std = 1/np.sqrt(len(pge_ends_comp))
    tval_crit = stats.t.ppf(1 - crit, df=len(pge_ends_comp) - 1)
    ax2.errorbar(1, pge_ends_cld.mean(), xerr=np.stack([[0.1, 0.1]]).T,
                 fmt='none', ecolor='C0')
    # Note: Because these are so close to zero, but are best plotted on a
    #        semilogy plot, the mean ± 2*σ/√N range or the mean ± Tc*σ/√N, where
    #        Tc is the critical t test value, can cross zero.
    ax2.errorbar(1, pge_ends_comp.mean(), yerr=pge_ends_comp.std() * tval_crit * scale_std,
                 fmt='oC1', elinewidth=20, ecolor='C1', alpha=0.5)
    # ax2.errorbar(0.5, pge_ends_comp.mean(), yerr=pge_ends_comp.std() * 2 * scale_std,
    #              fmt='k.', elinewidth=20, ecolor='C1', alpha=0.5)

    ax2.set_xlim([0.8, 1.2])
    ax2.set_xticks([1])
    ax2.set_xticklabels([args.variables[-1]], rotation=45, ha='right')

    plt.tight_layout()
    plt.savefig(img_file, bbox_inches='tight')
    plt.close(fig)

    img_desc = 'Left: The evolution of the maximum temperature (K) RMSE over a ' \
               'single time step for the {test} simulation (orange) and the {ref} ' \
               'simulation (blue), plotted with a log scale on the y-axis. ' \
               'The x-axis details the physical parameterizations ' \
               'and/or Fortran code modules executed within this time step. ' \
               'Right: the blue line indicates the {ref} ensemble mean at the ' \
               'end of the time step and the orange circle is the {test} ensemble mean. ' \
               'The orange box highlights the threshold values corresponding to the ' \
               'critical P {crit}% in the two-sided t-test. For the test to pass, ' \
               'the orange box must overlap the blue line. Note: The orange box may appear ' \
               'exceptionally large as  these  values  are very close to zero and ' \
               'the mean ± Tc*σ/√N range may cross zero, where Tc is the  critical ' \
               't-test value, σ is the ensemble standard deviation, N is the size ' \
               'of the ensemble, and σ/√N represents the t-test scaling ' \
               'parameter.'.format(test=args.test_name, ref=args.ref_name, crit=crit * 100)
    img_link = os.path.join(os.path.basename(args.img_dir), os.path.basename(img_file))
    img_gallery = el.gallery('', [
        el.image(args.test_case, img_desc, img_link, height=600)
    ])
    return details, img_gallery


def run(name, config, print_details=False):
    """
    Runs the extension.

    Args:
        name: The name of the extension
        config: The test's config dictionary
        print_details: Whether to print the analysis details to stdout
                       (default: False)

    Returns:
       A LIVVkit page element containing the LIVVkit elements to display on a webpage
    """

    # FIXME: move into a config to NameSpace function
    test_args = OrderedDict([(k.replace('-', '_'), v) for k, v in config.items()])
    test_args = argparse.Namespace(**test_args)

    test_args.img_dir = os.path.join(livvkit.output_dir, 'validation', 'imgs', name)
    fn.mkdir_p(test_args.img_dir)

    details, img_gal = main(test_args)

    tbl_el = {'Type': 'Table',
              'Title': 'Results',
              'Headers': ['Test status', 'Null hypothesis', 'T test (t, p)', 'Ensembles'],
              'Data': {'Null hypothesis': details['h0'],
                       'T test (t, p)': details['T test (t, p)'],
                       'Test status': 'pass' if details['h0'] == 'accept' else 'fail',
                       'Ensembles': 'statistically identical' if details['h0'] == 'accept' else 'statistically different'}
              }

    if print_details:
        _print_details(details)

    bib_html = bib2html(os.path.join(os.path.dirname(__file__), 'pg.bib'))
    tab_list = [el.tab('Figures', element_list=[img_gal]),
                el.tab('References', element_list=[el.html(bib_html)])]
    page = el.page(name, __doc__.replace('\n\n', '<br><br>'), element_list=[tbl_el], tab_list=tab_list)

    return page


def print_summary(summary):
    print('    Perturbation growth test: {}'.format(summary['']['Case']))
    print('      Null hypothesis: {}'.format(summary['']['Null hypothesis']))
    print('      T Test (t, p): {}'.format(summary['']['T test (t, p)']))
    print('      Ensembles: {}\n'.format(summary['']['Ensembles']))


def summarize_result(results_page):
    summary = {'Case': results_page['Title']}
    for elem in results_page['Data']['Elements']:
        if elem['Type'] == 'Table' and elem['Title'] == 'Results':
            summary['Test status'] = 'pass' if elem['Data']['Null hypothesis'] == 'accept' else 'fail'
            summary['Null hypothesis'] = elem['Data']['Null hypothesis']
            summary['T test (t, p)'] = elem['Data']['T test (t, p)']
            summary['Ensembles'] = 'statistically identical' if elem['Data']['Null hypothesis'] == 'accept' else 'statistically different'
            break
        else:
            continue
    return {'': summary}


def populate_metadata():
    """
    Generates the metadata needed for the output summary page
    """
    metadata = {'Type': 'ValSummary',
                'Title': 'Validation',
                'TableTitle': 'Perturbation growth test',
                'Headers': ['Test status', 'Null hypothesis', 'T test (t, p)', 'Ensembles']
                }
    return metadata


if __name__ == '__main__':
    test_name, test_config = parse_args()
    run(test_name, test_config, print_details=True)



#!/usr/bin/env python
# coding=utf-8
# Copyright (c) 2018 UT-BATTELLE, LLC
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""The Kolmogorov-Smirnov Test:
This tests the null hypothesis that the reference (n) and modified (m) model
Short Independent Simulation Ensembles (SISE) represent the same climate
state, based on the equality of distribution of each variable's annual global
average in the standard monthly model output between the two simulations.

The (per variable) null hypothesis uses the non-parametric, two-sample (n and m)
Kolmogorov-Smirnov test as the univariate test of equality of distribution of
global means. The test statistic (t) is the number of variables that reject the
(per variable) null hypothesis of equality of distribution at a 95% confidence
level. The (overall) null hypothesis is rejected if t > α, where α is some
critical number of rejecting variables. The critical value, α, is obtained from
an empirically derived approximate null distribution of t using resampling
techniques.
"""

from __future__ import absolute_import, division, print_function, unicode_literals
import six

import os
import argparse

from pprint import pprint
from collections import OrderedDict
    
import numpy as np
from scipy import stats

import livvkit
from livvkit.util import elements as el
from livvkit.util import functions as fn
from livvkit.util.LIVVDict import LIVVDict

from evv4esm.ensembles import e3sm
from evv4esm.ensembles.tools import monthly_to_annual_avg, prob_plot
from evv4esm.utils import bib2html


def variable_set(name):
    var_sets = fn.read_json(os.path.join(os.path.dirname(__file__),
                                         'ks_vars.json'))
    try:
        the_set = var_sets[name.lower()]
        return set(the_set)
    except KeyError as e:
        six.raise_from(argparse.ArgumentTypeError(
                'Unknown variable set! Known sets are {}'.format(
                        var_sets.keys()
                )), e)


def parse_args(args=None):
    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('-c', '--config',
                        type=fn.read_json,
                        help='A JSON config file containing a `ks` dictionary defining ' +
                             'the options. NOTE: command line options will override file options.')

    parser.add_argument('--test-case',
                        default='default',
                        help='Name of the test case.')

    parser.add_argument('--test-dir',
                        default=os.path.join(os.getcwd(), 'archive'),
                        help='Location of the test case run files.')

    parser.add_argument('--ref-case',
                        default='fast',
                        help='Name of the reference case.')

    parser.add_argument('--ref-dir',
                        default=os.path.join(os.getcwd(), 'archive'),
                        help='Location of the reference case run files.')

    parser.add_argument('--var-set',
                        default='default', type=variable_set,
                        help='Name of the variable set to analyze.')

    parser.add_argument('--ninst',
                        default=30, type=int,
                        help='The number of instances (should be the same for '
                             'both cases).')

    parser.add_argument('--critical',
                        default=13, type=float,
                        help='The critical value (desired significance level) for rejecting the ' +
                        'null hypothesis.')
   
    parser.add_argument('--img-dir',
                        default=os.getcwd(),
                        help='Image output location.')

    args, _ = parser.parse_known_args(args)

    # use config file arguments, but override with command line arguments
    if args.config:
        default_args = parser.parse_args([])

        for key, val, in vars(args).items():
            if val != vars(default_args)[key]:
                args.config['ks'][key] = val

        config_arg_list = []
        [config_arg_list.extend(['--'+key, str(val)]) for key, val in args.config['ks'].items()
         if key != 'config']
        args, _ = parser.parse_known_args(config_arg_list)

    return args


def run(name, config):
    """
    Runs the analysis.

    Args:
        name: The name of the test
        config: A dictionary representation of the configuration file

    Returns:
       The result of elements.page with the list of elements to display
    """
   
    config_arg_list = []
    [config_arg_list.extend(['--'+key, str(val)]) for key, val in config.items()]

    args = parse_args(config_arg_list)

    args.img_dir = os.path.join(livvkit.output_dir, 'validation', 'imgs', name)
    fn.mkdir_p(args.img_dir)

    details, img_gal = main(args)

    tbl_data = OrderedDict(sorted(details.items()))
    tbl_el = {'Type': 'V-H Table',
              'Title': 'Validation',
              'TableTitle': 'Analyzed variables',
              'Headers': ['h0', 'K-S test (D, p)', 'T test (t, p)'],
              'Data': {'': tbl_data}
              }
    bib_html = bib2html(os.path.join(os.path.dirname(__file__), 'ks.bib'))
    tl = [el.tab('Figures', element_list=[img_gal]),
          el.tab('Details', element_list=[tbl_el]),
          el.tab('References', element_list=[el.html(bib_html)])]

    rejects = [var for var, dat in tbl_data.items() if dat['h0'] == 'reject']
    results = {'Type': 'Table',
               'Title': 'Results',
               'Headers': ['Test status', 'Variables analyzed', 'Rejecting', 'Critical value', 'Ensembles'],
               'Data': {'Test status': 'pass' if len(rejects) < args.critical else 'fail',
                        'Variables analyzed': len(tbl_data.keys()),
                        'Rejecting': len(rejects),
                        'Critical value': args.critical,
                        'Ensembles': 'statistically identical' if len(rejects) < args.critical else 'statistically different'}
               }

    # FIXME: Put into a ___ function
    page = el.page(name, __doc__.replace('\n\n', '<br><br>'), element_list=[results], tab_list=tl)
    return page


def case_files(args):
    # ensure unique case names for the dictionary
    key1 = args.test_case
    key2 = args.ref_case
    if args.test_case == args.ref_case:
        key1 += '1'
        key2 += '2'

    f_sets = {key1: e3sm.component_monthly_files(args.test_dir, 'cam', args.ninst),
              key2: e3sm.component_monthly_files(args.ref_dir, 'cam', args.ninst)}

    return f_sets, key1, key2


def print_summary(summary):
    print('    Kolmogorov-Smirnov Test: {}'.format(summary['']['Case']))
    print('      Variables analyzed: {}'.format(summary['']['Variables analyzed']))
    print('      Rejecting: {}'.format(summary['']['Rejecting']))
    print('      Critical value: {}'.format(summary['']['Critical value']))
    print('      Ensembles: {}'.format(summary['']['Ensembles']))
    print('      Test status: {}\n'.format(summary['']['Test status']))


def print_details(details):
    for set_ in details:
        print('-'*80)
        print(set_)
        print('-'*80)
        pprint(details[set_])


def summarize_result(results_page):
    summary = {'Case': results_page['Title']}
    for elem in results_page['Data']['Elements']:
        if elem['Type'] == 'Table' and elem['Title'] == 'Results':
            summary['Test status'] = elem['Data']['Test status']
            summary['Variables analyzed'] = elem['Data']['Variables analyzed']
            summary['Rejecting'] = elem['Data']['Rejecting']
            summary['Critical value'] = elem['Data']['Critical value']
            summary['Ensembles'] = elem['Data']['Ensembles']
            break
        else:
            continue
    return {'': summary}


def populate_metadata():
    """
    Generates the metadata responsible for telling the summary what
    is done by this module's run method
    """
    
    metadata = {'Type': 'ValSummary',
                'Title': 'Validation',
                'TableTitle': 'Kolmogorov-Smirnov test',
                'Headers': ['Test status', 'Variables analyzed', 'Rejecting', 'Critical value', 'Ensembles']}
    return metadata
    

def main(args):
    ens_files, key1, key2 = case_files(args)
    if args.test_case == args.ref_case:
        args.test_case = key1
        args.ref_case = key2

    monthly_avgs = e3sm.gather_monthly_averages(ens_files, args.var_set)
    annual_avgs = monthly_avgs.groupby(['case', 'variable', 'instance']
                                       ).monthly_mean.aggregate(monthly_to_annual_avg).reset_index()

    # now, we got the data, so let's get some stats
    test_set = set(monthly_avgs[monthly_avgs.case == args.test_case].variable.unique())
    ref_set = set(monthly_avgs[monthly_avgs.case == args.ref_case].variable.unique())
    common_vars = list(test_set & ref_set)

    img_list = []
    details = LIVVDict()
    for var in sorted(common_vars):
        annuals_1 = annual_avgs.query('case == @args.test_case & variable == @var').monthly_mean.values
        annuals_2 = annual_avgs.query('case == @args.ref_case & variable == @var').monthly_mean.values

        details[var]['T test (t, p)'] = stats.ttest_ind(annuals_1, annuals_2,
                                                        equal_var=False, nan_policy=str('omit'))
        if np.isnan(details[var]['T test (t, p)']).any() or np.isinf(details[var]['T test (t, p)']).any():
            details[var]['T test (t, p)'] = (None, None)

        details[var]['K-S test (D, p)'] = stats.ks_2samp(annuals_1, annuals_2)

        details[var]['mean (test case, ref. case)'] = (annuals_1.mean(), annuals_2.mean())

        details[var]['max (test case, ref. case)'] = (annuals_1.max(), annuals_2.max())

        details[var]['min (test case, ref. case)'] = (annuals_1.min(), annuals_2.min())

        details[var]['std (test case, ref. case)'] = (annuals_1.std(), annuals_2.std())

        if details[var]['T test (t, p)'][0] is None:
            details[var]['h0'] = '-'
        elif details[var]['K-S test (D, p)'][1] < 0.05:
            details[var]['h0'] = 'reject'
        else:
            details[var]['h0'] = 'accept'

        img_file = os.path.relpath(os.path.join(args.img_dir, var + '.png'), os.getcwd())
        prob_plot(annuals_1, annuals_2, 20, img_file, test_name=args.test_case, ref_name=args.ref_case,
                  pf=details[var]['h0'])
        
        img_desc = 'Mean annual global average of {} for <em>{}</em> is {:.3e} ' \
                   'and for <em>{}</em> is {:.3e}'.format(var,
                                                          args.test_case,
                                                          details[var]['mean (test case, ref. case)'][0],
                                                          args.ref_case,
                                                          details[var]['mean (test case, ref. case)'][1])

        img_link = os.path.join(os.path.basename(args.img_dir), os.path.basename(img_file))
        img_list.append(el.image(var, img_desc, img_link))
        
    img_gal = el.gallery('Analyzed variables', img_list)

    return details, img_gal


if __name__ == '__main__':
    print_details(main(parse_args()))



#!/usr/bin/env python
# coding=utf-8
# Copyright (c) 2018 UT-BATTELLE, LLC
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""E3SM specific ensemble functions."""

from __future__ import absolute_import, division, print_function, unicode_literals
import six

import os
import re
import glob

from collections import OrderedDict

import numpy as np
import pandas as pd
from netCDF4 import Dataset


def component_file_instance(component, case_file):
    search_regex = r'{c}_[0-9]+'.format(c=component)
    result = re.search(search_regex, case_file).group(0)
    return int(result.replace('cam_', ''))


def file_date_str(case_file, style='short'):
    if style == 'full':
        search_regex = r'h0\.[0-9]+-[0-9]+-[0-9]+-[0-9]+.nc'
    elif style == 'short':
        search_regex = r'h0\.[0-9]+-[0-9]+.nc'
    else:
        # FIXME: log warning here
        search_regex = r'h0\.[0-9]+-[0-9]+.nc'
    result = re.search(search_regex, case_file).group(0)
    return result.replace('h0.', '').replace('.nc', '')


def component_monthly_files(dir_, component, ninst):
    base = '{d}/*{c}_????.h0.????-??.nc'.format(d=dir_, c=component)
    search = os.path.normpath(base)
    result = sorted(glob.glob(search))

    instance_files = OrderedDict()
    for ii in range(1, ninst + 1):
        instance_files[ii] = sorted(filter(lambda x: component_file_instance(component, x) == ii, result),
                                    key=file_date_str)
        if len(instance_files[ii]) > 12:
            instance_files[ii] = instance_files[ii][-12:]

    return instance_files


def gather_monthly_averages(ensemble_files, variable_set=None):
    monthly_avgs = []
    for case, inst_dict in six.iteritems(ensemble_files):
        for inst, i_files in six.iteritems(inst_dict):
            # Get monthly averages from files
            for file_ in i_files:
                date_str = file_date_str(file_)

                data = None
                try:
                    data = Dataset(file_)
                    if variable_set is None:
                        variable_set = set(data.variables.keys())
                except OSError as E:
                    six.raise_from(BaseException('Could not open netCDF dataset: {}'.format(file_)), E)

                for var in data.variables.keys():
                    if var not in variable_set:
                        continue
                    if len(data.variables[var].shape) < 2 or var in ['time_bnds', 'date_written', 'time_written']:
                        continue
                    elif 'ncol' not in data.variables[var].dimensions:
                        continue
                    else:
                        m = np.mean(data.variables[var][0, ...])

                    monthly_avgs.append((case, var, '{:04}'.format(inst), date_str, m))

    monthly_avgs = pd.DataFrame(monthly_avgs, columns=('case', 'variable', 'instance', 'date', 'monthly_mean'))
    return monthly_avgs



# coding=utf-8

from __future__ import absolute_import, print_function, unicode_literals
import six

import pybtex.database
import pybtex.io

from pybtex.backends.html import Backend as BaseBackend
from pybtex.style.formatting.plain import Style as PlainStyle


class HTMLBackend(BaseBackend):
    def __init__(self, *args, **kwargs):
        super(HTMLBackend, self).__init__(*args, **kwargs)
        self._html = ''

    def output(self, html):
        self._html += html

    def format_protected(self, text):
        if text[:4] == 'http':
            return self.format_href(text, text)
        else:
            return r'<span class="bibtex-protected">{}</span>'.format(text)

    def write_prologue(self):
        self.output('<div class="bibliography"><dl>')

    def write_epilogue(self):
        self.output('</dl></div>')

    def _repr_html(self, formatted_bibliography):
        self.write_prologue()
        for entry in formatted_bibliography:
            self.write_entry(entry.key, entry.label, entry.text.render(self))
        self.write_epilogue()

        return self._html.replace('\n', ' ').replace('\\url <a', '<a')


# FIXME: For python 3.7+ only...
# from functools import singledispatch
# from collections.abc import Iterable
#
# # noinspection PyUnusedLocal
# @singledispatch
# def bib2html(bib, style=None, backend=None):
#     raise NotImplementedError('I do not now how to convert a {} type to a bibliography'.format(type(bib)))
def bib2html(bib, style=None, backend=None):
    if isinstance(bib, six.string_types):
        return _bib2html_string(bib, style=style, backend=backend)
    if isinstance(bib, (list, set, tuple)):
        return _bib2html_list(bib, style=style, backend=backend)
    if isinstance(bib, pybtex.database.BibliographyData):
        return _bib2html_bibdata(bib, style=style, backend=backend)
    else:
        raise NotImplementedError('I do not now how to convert a {} type to a bibliography'.format(type(bib)))


# FIXME: For python 3.7+ only...
# @bib2html.register
# def _bib2html_string(bib: str, style=None, backend=None):
def _bib2html_string(bib, style=None, backend=None):
    if style is None:
        style = PlainStyle()
    if backend is None:
        backend = HTMLBackend()

    formatted_bib = style.format_bibliography(pybtex.database.parse_file(bib))

    return backend._repr_html(formatted_bib)


# FIXME: For python 3.7+ only...
# @bib2html.register
# def _bib2html_list(bib: Iterable, style=None, backend=None):
def _bib2html_list(bib, style=None, backend=None):
    if style is None:
        style = PlainStyle()
    if backend is None:
        backend = HTMLBackend()

    bibliography = pybtex.database.BibliographyData()
    for bib_file in bib:
        temp_bib = pybtex.database.parse_file(bib_file)
        for key, entry in temp_bib.entries.items():
            try:
                bibliography.add_entry(key, entry)
            except pybtex.database.BibliographyDataError:
                continue

    formatted_bib = style.format_bibliography(bibliography)

    return backend._repr_html(formatted_bib)


# FIXME: For python 3.7+ only...
# @bib2html.register
# def _bib2html_bibdata(bib: pybtex.database.BibliographyData, style=None, backend=None):
def _bib2html_bibdata(bib, style=None, backend=None):
    if style is None:
        style = PlainStyle()
    if backend is None:
        backend = HTMLBackend()

    formatted_bib = style.format_bibliography(bib)

    return backend._repr_html(formatted_bib)



# coding=utf-8
# Copyright (c) 2018 UT-BATTELLE, LLC
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import, print_function, unicode_literals

import os
import re

from setuptools import setup


here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.md'), 'r') as f:
    long_desc = f.read()
with open(os.path.join(here, 'evv4esm', '__init__.py')) as f:
    init_file = f.read()

setup(
      name='evv4esm',
      version=re.search(r'{}\s*=\s*[(]([^)]*)[)]'.format('__version_info__'),
                        init_file
                        ).group(1).replace(', ', '.'),

      description='Extended verification and validation for earth system models',
      long_description=long_desc,
      long_description_content_type='text/markdown',

      url='https://github.com/LIVVkit/evv4esm',

      author='Joseph H. Kennedy',
      author_email='kennedyjh@ornl.gov',

      license='BSD',
      include_package_data=True,

      classifiers=['Development Status :: 5 - Production/Stable',

                   'Intended Audience :: Science/Research',
                   'Topic :: Scientific/Engineering',
                   'Topic :: Software Development :: Testing',

                   'License :: OSI Approved :: BSD License',

                   'Programming Language :: Python :: 2',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3',
                   'Programming Language :: Python :: 3.5',
                   'Programming Language :: Python :: 3.6',
                   ],

      install_requires=[
          'six',
          'numpy',
          'scipy',
          'pandas',
          'livvkit',
          'netCDF4',
          'matplotlib',
          'pybtex',
          ],

      packages=[
          'evv4esm'
          ],

      entry_points={'console_scripts': ['evv = evv4esm.__main__:main']},

      zip_safe=False,
      )
